# Copyright (C) 2019 Catalyst Cloud Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from confspirator import base
from confspirator import exceptions


def recursive_merge(overridden, overrider):
    """Recursively merges a dict or a group over another group

    :param overridden: a dict or group to override
    :param overrider: a dict or group to override the other with

    If both overridden and overrider have a key who's value is a dict
    or group then recursive_merge is called on both values and the result
    stored in the returned dictionary.
    """
    if not isinstance(overrider, (base.BaseGroupNamespace, dict)):
        return overrider

    result = copy.deepcopy(overridden)
    if isinstance(overrider, base.BaseGroupNamespace):
        items = overrider._values.items()
    else:
        items = overrider.items()

    for k, v in items:
        if k in result and isinstance(result[k], (dict, base.BaseGroupNamespace)):
            result_value = recursive_merge(result[k], v)
        else:
            result_value = copy.deepcopy(v)

        if isinstance(result, base.BaseGroupNamespace):
            result._values[k] = result_value
        else:
            result[k] = result_value
    return result


def validate_config_format(config_format):
    if config_format in ["yml", "yaml", "YML", "YAML"]:
        return "yaml"
    if config_format in ["tml", "toml", "TML", "TOML"]:
        return "toml"
    raise exceptions.InvalidConfigFileFormat(config_format)



# Copyright (C) 2019 Catalyst Cloud Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from slugify import slugify

from confspirator import exceptions
from confspirator import constants
from confspirator import types


class BaseConfigField(object):

    """Base class for all configuration options.
    The only required parameter is the config's name. However, it is
    common to also supply a default and help string for all options.

    :param name: the config's name
    :param config_type: the option's type. Must be a callable object that
                        takes string and returns converted and validated value
    :param help_text: an explanation of how the option is used
    :param default: the default value of the option
    :param required: true if a value must be supplied for this option
    :param test_default: a default for when running in test mode
    :param sample_default: a default for sample config files
    :param unsafe_default: true if the default value must be overridden
    :param secret: true if the value should be obfuscated in log output
    :param deprecated_location: deprecated dot separated location.
                                e.g. 'service.group1.old_name'
                                Acts like an alias.
    :param deprecated_for_removal: indicates whether this opt is planned for
                                   removal in a future release
    :param deprecated_reason: indicates why this opt is planned for removal in
                              a future release. Silently ignored if
                              deprecated_for_removal is False
    :param deprecated_since: indicates which release this opt was deprecated
                             in. Accepts any string, though valid version
                             strings are encouraged. Silently ignored if
                             deprecated_for_removal is False
    :param advanced: a bool True/False value if this option has advanced usage
                             and is not normally used by the majority of users
    """

    def __init__(
        self,
        name,
        config_type=None,
        help_text=None,
        default=None,
        test_default=None,
        sample_default=None,
        unsafe_default=False,
        secret=False,
        required=False,
        required_for_tests=None,
        deprecated_location=None,
        deprecated_for_removal=False,
        deprecated_reason=None,
        deprecated_since=None,
        advanced=False,
        reformat_name=True,
    ):
        if reformat_name:
            name = slugify(name)
        # we still at the very least do some basic formatting
        name = name.replace("-", "_")
        if name.startswith("_"):
            raise ValueError("illegal name %s with prefix _" % (name,))
        self.name = name

        if config_type is None:
            config_type = types.String()

        if not callable(config_type):
            raise TypeError("type must be callable")
        self.type = config_type

        self.help_text = help_text
        self.default = default
        self.test_default = test_default
        self.sample_default = sample_default
        self.unsafe_default = unsafe_default
        self.secret = secret
        self.required = required
        if required_for_tests is None:
            self.required_for_tests = required
        else:
            self.required_for_tests = required_for_tests
        self.deprecated_location = deprecated_location
        self.deprecated_for_removal = deprecated_for_removal
        self.deprecated_reason = deprecated_reason
        self.deprecated_since = deprecated_since
        self.advanced = advanced

        self._group = None

    def set_parent(self, parent):
        self._group = parent

    def path(self, deprecated=False):
        if deprecated:
            return self.deprecated_location.split(".")
        return self._group.path() + [self.name]

    def path_str(self):
        return ".".join(self.path())

    def envvar_name(self, deprecated=False):
        """Make the expected envvar key

        This is in the style of 'GROUPNAME_SUBGROUPNAME_KEY'
        and is generated automatically based on the path into
        the conf.
        """
        return "_".join(self.path(deprecated)).upper()

    def __str__(self):
        if self._group:
            return self.path_str()
        else:
            return self.name

    def parse_value(self, value, test_mode):
        conf_name = ".".join(self.path())
        errors = []
        if value == constants.DEFAULT_NONE_VALUE:
            if self.unsafe_default and not test_mode:
                print(
                    "WARNING(%s): "
                    "This conf has an unsafe default you must override." % conf_name
                )

            if test_mode:
                if self.required and (
                    self.default is None and self.test_default is None
                ):
                    if self.required_for_tests:
                        errors.append("This conf is required.")
                if self.test_default is not None:
                    value = self.test_default
                else:
                    value = self.default
            else:
                if self.required and self.default is None:
                    errors.append("This conf is required.")
                value = self.default

        if self.deprecated_for_removal:
            print("WARNING(%s): " "This conf has been deprecated." % conf_name)

        if value is not None:
            try:
                return self.type(value)
            except ValueError as e:
                errors.append(str(e))

        if errors:
            raise exceptions.InvalidConfValue(errors)


class StrConfig(BaseConfigField):
    """Config with String type

    :param name: the config's name
    :param choices: Optional sequence of either valid values or tuples of valid
        values with descriptions.
    :param quotes: If True and string is enclosed with single or double
                   quotes, will strip those quotes.
    :param regex: Optional regular expression (string or compiled
                  regex) that the value must match on an unanchored
                  search.
    :param ignore_case: If True case differences (uppercase vs. lowercase)
                        between 'choices' or 'regex' will be ignored.
    :param max_length: If positive integer, the value must be less than or
                       equal to this parameter.
    """

    def __init__(
        self,
        name,
        choices=None,
        quotes=None,
        regex=None,
        ignore_case=False,
        max_length=None,
        **kwargs
    ):
        super(StrConfig, self).__init__(
            name,
            config_type=types.String(
                choices=choices,
                quotes=quotes,
                regex=regex,
                ignore_case=ignore_case,
                max_length=max_length,
            ),
            **kwargs
        )


class BoolConfig(BaseConfigField):
    """Boolean config

    :param name: the config's name
    """

    def __init__(self, name, **kwargs):
        super(BoolConfig, self).__init__(name, config_type=types.Boolean(), **kwargs)


class IntConfig(BaseConfigField):
    """Config with Integer type

    :param name: the config's name
    :param min: minimum value the integer can take
    :param max: maximum value the integer can take
    """

    def __init__(self, name, min=None, max=None, **kwargs):
        super(IntConfig, self).__init__(
            name, config_type=types.Integer(min, max), **kwargs
        )


class FloatConfig(BaseConfigField):
    """Config with Float type

    :param name: the config's name
    :param min: minimum value the float can take
    :param max: maximum value the float can take
    """

    def __init__(self, name, min=None, max=None, **kwargs):
        super(FloatConfig, self).__init__(
            name, config_type=types.Float(min, max), **kwargs
        )


class ListConfig(BaseConfigField):
    """Option with List(String) type

    :param name: the config's name
    :param item_type: type of items (see :class:`confspirator.types`)
    """

    def __init__(self, name, item_type=None, **kwargs):
        super(ListConfig, self).__init__(
            name, config_type=types.List(item_type=item_type), **kwargs
        )


class DictConfig(BaseConfigField):
    """Config with Dict(String) type

    :param name: the config's name
    :param value_type: type of values (see :class:`confspirator.types`)
    :param check_value_type: if value is already dict, should we check
                              value type
    :param is_json: if True and value is string, will parse as json
    """

    def __init__(
        self, name, value_type=None, check_value_type=False, is_json=False, **kwargs
    ):
        super(DictConfig, self).__init__(
            name,
            config_type=types.Dict(
                value_type=value_type,
                check_value_type=check_value_type,
                is_json=is_json,
            ),
            **kwargs
        )


class IPConfig(BaseConfigField):
    """Config with IPAddress type

    :param name: the config's name
    :param version: one of either ``4``, ``6``, or ``None`` to specify
       either version.
    """

    def __init__(self, name, version=None, **kwargs):
        super(IPConfig, self).__init__(
            name, config_type=types.IPAddress(version), **kwargs
        )


class PortConfig(BaseConfigField):
    """Config for a TCP/IP port number.  Ports can range from 0 to 65535.

    :param name: the config's name
    :param min: minimum value the port can take
    :param max: maximum value the port can take
    :param choices: Sequence of valid values.
    """

    def __init__(self, name, min=None, max=None, choices=None, **kwargs):
        type = types.Port(min=min, max=max, choices=choices, type_name="port value")
        super(PortConfig, self).__init__(name, config_type=type, **kwargs)


class HostNameConfig(BaseConfigField):
    """Config for a hostname.  Only accepts valid hostnames.

    :param name: the config's name
    """

    def __init__(self, name, **kwargs):
        super(HostNameConfig, self).__init__(
            name, config_type=types.Hostname(), **kwargs
        )


class HostAddressConfig(BaseConfigField):
    """Option for either an IP or a hostname.

    Accepts valid hostnames and valid IP addresses.

    :param name: the config's name
    :param version: one of either ``4``, ``6``, or ``None`` to specify
       either version.
    """

    def __init__(self, name, version=None, **kwargs):
        super(HostAddressConfig, self).__init__(
            name, config_type=types.HostAddress(version), **kwargs
        )


class URIConfig(BaseConfigField):
    """Opt with URI type

    :param name: the config's name
    :param max_length: If positive integer, the value must be less than or
                       equal to this parameter.
    :param schemes: list of valid URI schemes, e.g. 'https', 'ftp', 'git'
    """

    def __init__(self, name, max_length=None, schemes=None, **kwargs):
        type = types.URI(max_length=max_length, schemes=schemes)
        super(URIConfig, self).__init__(name, config_type=type, **kwargs)



# Copyright (C) 2019 Catalyst Cloud Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import copy
from functools import WRAPPER_ASSIGNMENTS, wraps
from unittest import TestCase

from confspirator import exceptions
from confspirator import base
from confspirator import utils


class modify_conf(object):
    """A test decorator to allow modifying the conf for tests

    Operations avaiable:
        value:
            - override
        list:
            - override
            - preprend
            - append
            - remove
        dict:
            - override
            - update
            - delete
            - overlay
        GroupNamespace:
            - override
            - overlay

    example usage:

    @modify_conf(
        conf=CONF,
        operations={
            "test_app.api_settings.item_list_options": [
                {"operation": "remove", "value": "option1"},
            ],
    })
    class MyTestClass(TestCase):

    :param conf: The root GroupNamespace
    :param operations: a dict of dot seperate paths to lists of operations
    """

    enable_exception = None

    def __init__(self, conf, operations):
        self.conf = conf
        self.operations = operations

    def __enter__(self):
        self.enable()

    def __exit__(self, exc_type, exc_value, traceback):
        self.disable()

    def decorate_class(self, cls):
        if issubclass(cls, TestCase):
            decorated_setUp = cls.setUp
            decorated_tearDown = cls.tearDown

            def setUp(inner_self):
                self.enable()
                decorated_setUp(inner_self)

            def tearDown(inner_self):
                decorated_tearDown(inner_self)
                self.disable()

            cls.setUp = setUp
            cls.tearDown = tearDown
            return cls
        raise TypeError("Can only decorate subclasses of unittest.TestCase")

    def decorate_callable(self, func):
        @wraps(func, assigned=WRAPPER_ASSIGNMENTS)
        def inner(*args, **kwargs):
            with self:
                return func(*args, **kwargs)

        return inner

    def __call__(self, decorated):
        if isinstance(decorated, type):
            return self.decorate_class(decorated)
        elif callable(decorated):
            return self.decorate_callable(decorated)
        raise TypeError("Cannot decorate object of type %s" % type(decorated))

    def _get_nested_value(self, conf, path):
        if len(path) == 1:
            return conf[path[0]]
        elif path:
            return self._get_nested_value(conf[path[0]], path[1:])
        return conf

    def _do_operation(self, root_key, path, operation):
        keys = path.split(".")
        if keys[0] != root_key:
            raise exceptions.NoSuchConfig(path)

        end_key = keys[-1]
        if operation["operation"] == "override":
            if len(keys) == 1:
                raise exceptions.InvalidTestOperation("Cannot override root conf")
            holder = self._get_nested_value(self.conf, keys[1:-1])
            if not isinstance(holder, base.BaseGroupNamespace):
                raise exceptions.InvalidTestOperation(
                    "Overridden key must have a conf group as parent."
                )
            holder._values[end_key] = operation["value"]
        elif operation["operation"] in ["prepend", "append", "remove"]:
            list_val = self._get_nested_value(self.conf, keys[1:])
            if not isinstance(list_val, list):
                raise exceptions.InvalidTestOperation(
                    "'%s' is a list only operation" % operation["operation"]
                )
            if operation["operation"] == "prepend":
                list_val.insert(0, operation["value"])
            if operation["operation"] == "append":
                list_val.append(operation["value"])
            if operation["operation"] == "remove":
                list_val.remove(operation["value"])
        elif operation["operation"] in ["update", "delete"]:
            dict_val = self._get_nested_value(self.conf, keys[1:])
            if not isinstance(dict_val, dict):
                raise exceptions.InvalidTestOperation(
                    "'%s' is a dict only operation" % operation["operation"]
                )
            if operation["operation"] == "update":
                dict_val.update(operation["value"])
            if operation["operation"] == "delete":
                del dict_val[operation["value"]]
        elif operation["operation"] == "overlay":
            dict_like_val = self._get_nested_value(self.conf, keys[1:])
            if not isinstance(dict_like_val, (dict, base.BaseGroupNamespace)):
                raise exceptions.InvalidTestOperation(
                    "'%s' is a dict only operation" % operation["operation"]
                )
            holder = self._get_nested_value(self.conf, keys[1:-1])
            if isinstance(holder, base.BaseGroupNamespace):
                holder._values[end_key] = utils.recursive_merge(
                    dict_like_val, operation["value"]
                )
            else:
                holder[end_key] = utils.recursive_merge(
                    dict_like_val, operation["value"]
                )

    def enable(self):
        self.wrapped = copy.deepcopy(self.conf._values)

        root_key = self.conf.name

        for path, operations in self.operations.items():
            try:
                for operation in operations:
                    self._do_operation(root_key, path, operation)
            except Exception as exc:
                self.enable_exception = exc
                self.disable()

    def disable(self):
        self.conf._values = self.wrapped
        del self.wrapped
        if self.enable_exception is not None:
            exc = self.enable_exception
            self.enable_exception = None
            raise exc



import toml
import yaml

from confspirator import exceptions
from confspirator import groups
from confspirator import utils


class YamlDumper(yaml.Dumper):
    """Custom dumper to deal with a weird list indentation issue.

    Remove when https://github.com/yaml/pyyaml/issues/234 is
    solved.
    """

    def increase_indent(self, flow=False, indentless=False):
        return super(YamlDumper, self).increase_indent(flow, False)


def _make_yaml_lines(field, default, depth, comment=False):
    new_lines = []
    line_prefix = "  " * depth

    field_type = field.type.__class__.__name__
    if field_type in ["Dict", "List"]:
        new_lines.append(f"{line_prefix}{field.name}:")
        line_prefix += "  "
        val = default
    else:
        val = {field.name: default}
    for line in yaml.dump(val, Dumper=YamlDumper, default_flow_style=False).split("\n"):
        if line == "":
            continue
        if comment:
            new_lines.append(f"{line_prefix}# {line}")
        else:
            new_lines.append(line_prefix + line)
    return new_lines


def _make_toml_lines(field, default, depth, comment=False):
    new_lines = []
    line_prefix = "  " * depth

    field_type = field.type.__class__.__name__
    if field_type == "Dict":
        new_lines.append(f"{line_prefix}[{field.path_str()}]")
        line_prefix += "  "

        val = {}
        pointer = val
        for name in field.path()[:-1]:
            new_dict = {}
            pointer[name] = new_dict
            pointer = new_dict
        pointer[field.name] = default
    else:
        val = {field.name: default}

    for line in toml.dumps(val).split("\n"):
        if line == "" or f"[{field.path_str()}]" in line:
            continue
        if comment:
            new_lines.append(f"{line_prefix}# {line}")
        else:
            new_lines.append(line_prefix + line)
    return new_lines


def _make_field_lines(field, depth, output_format):
    field_lines = []
    line_prefix = "  " * depth

    field_type = field.type.__class__.__name__
    if field.help_text:
        field_help_text = f"{line_prefix}# {field_type} - {field.help_text}"
    else:
        field_help_text = f"{line_prefix}# {field_type}"
    field_lines.append(field_help_text)

    if field.deprecated_for_removal:
        deprecated_text = f"{line_prefix}# DEPRECATED"
        if field.deprecated_reason:
            deprecated_text += f" - {field.deprecated_reason}"
        if field.deprecated_since:
            deprecated_text += f" - {field.deprecated_since}"
        field_lines.append(deprecated_text)

    default = ""
    if field.default is not None:
        default = field.default

    if not default and field.sample_default is not None:
        default = field.sample_default

    if field_type in ["Dict", "List"]:
        if default:
            if output_format == "yaml":
                field_lines += _make_yaml_lines(field, default, depth)
            elif output_format == "toml":
                field_lines += _make_toml_lines(field, default, depth)
        else:
            if output_format == "yaml":
                field_lines.append(f"{line_prefix}# {field.name}:")
            elif output_format == "toml":
                if field_type == "List":
                    field_lines.append(f"{line_prefix}# {field.name} = []")
                elif field_type == "Dict":
                    field_lines.append(f"{line_prefix}# [{field.path_str()}]")
    else:
        if default == "":
            if output_format == "yaml":
                field_lines.append(f"{line_prefix}# {field.name}: <your_value>")
            elif output_format == "toml":
                field_lines.append(f"{line_prefix}# {field.name} = <your_value>")
        else:
            if output_format == "yaml":
                field_lines += _make_yaml_lines(field, default, depth)
            elif output_format == "toml":
                field_lines += _make_toml_lines(field, default, depth)
    return field_lines


def _make_group_lines(group, output_format, depth=0):
    group_lines = []
    line_prefix = "  " * depth

    if group.description:
        group_lines.append(f"{line_prefix}# {group.description}")

    if output_format == "yaml":
        group_lines.append(f"{line_prefix}{group.name}:")
    elif output_format == "toml":
        group_lines.append(f"{line_prefix}[{ '.'.join(group.path()) }]")

    for child in group:
        if isinstance(child, groups.ConfigGroup):
            group_lines += _make_group_lines(child, output_format, depth=depth + 1)
        else:
            group_lines += _make_field_lines(child, depth + 1, output_format)
    return group_lines


def _create_example_lines(config_group, output_format):
    config_lines = []
    top_level_fields = False
    for child in config_group:
        if isinstance(child, groups.ConfigGroup):
            config_lines += _make_group_lines(child, output_format)
        else:
            top_level_fields = True
            config_lines += _make_field_lines(child, 0, output_format)
        # all top level configs or groups should be spaced
        config_lines.append("")
    if output_format == "toml" and top_level_fields:
        config_lines.insert(0, f"[{config_group.name}]")
    return config_lines


def create_example_config(config_group, output_file, output_format=None):
    if not isinstance(config_group, groups.ConfigGroup):
        raise exceptions.InvalidConfigClass(
            f"'{config_group}' is not a valid config class"
        )

    if not output_format:
        output_format = utils.validate_config_format(output_file.split(".")[-1])
    else:
        output_format = utils.validate_config_format(output_format)

    config_lines = _create_example_lines(config_group, output_format)

    with open(output_file, "w") as f:
        for line in config_lines:
            f.write(line)
            f.write("\n")



# Copyright (C) 2019 Catalyst Cloud Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


class InvalidConfigClass(Exception):
    pass


class ConfigGroupAlreadyRegistered(Exception):
    pass


class InvalidConfigFileFormat(Exception):
    def __init__(self, config_format=None):
        self.msg = "Cannot use unsupported config formats."
        if config_format:
            self.msg = f"'{config_format}' is not a supported config format."

    def __str__(self):
        return str(self.msg)


class BaseConfigLoaderException(Exception):
    """ """

    def __init__(self, errors=None):
        self.errors = errors

    def __str__(self):
        return str(self.errors)


class InvalidConfValue(BaseConfigLoaderException):
    pass


class InvalidConfGroup(BaseConfigLoaderException):
    pass


class InvalidConf(BaseConfigLoaderException):
    pass


class NoSuchConfig(KeyError):
    pass


class InvalidTestOperation(Exception):
    pass



# Copyright (C) 2019 Catalyst Cloud Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

import toml
import yaml

from confspirator import constants
from confspirator import exceptions
from confspirator import base
from confspirator import groups
from confspirator import utils


def _get_nested_value(conf, path):
    """Simple recursive function to get a value down a path in a dict"""
    if conf is None:
        raise KeyError("No value in config dict.")
    if len(path) == 1:
        return conf[path[0]]
    else:
        return _get_nested_value(conf[path[0]], path[1:])


def _get_value(conf, field):
    """Get the given config field value from the conf dict

    Will attempt to get the key at the given path, or fallback
    to a deprecated path if specified.
    """
    path = field.path()
    try:
        return _get_nested_value(conf, path)
    except KeyError:
        envvar_name = field.envvar_name()
        result = os.environ.get(envvar_name, constants.DEFAULT_NONE_VALUE)

    if result == constants.DEFAULT_NONE_VALUE and field.deprecated_location:
        path = field.path(deprecated=True)
        try:
            return _get_nested_value(conf, path)
        except KeyError:
            envvar_name = field.envvar_name(deprecated=True)
            result = os.environ.get(envvar_name, constants.DEFAULT_NONE_VALUE)
    return result


def process_group(group, conf, test_mode, lazy_loading=False):
    """Process a given ConfigGroup against the conf dict into a GroupNamespace"""
    if not isinstance(group, base.BaseConfigGroup):
        raise exceptions.InvalidConfGroup("'%s' is not a valid ConfigGroup." % group)

    if group.lazy_load and not lazy_loading:
        return groups.LazyLoadedGroupNamespace(group.name, group, conf, test_mode)

    parsed_children = {}
    errors = {}
    for child in group:
        if isinstance(child, base.BaseConfigGroup):
            parsed_children[child.name] = process_group(child, conf, test_mode)
            continue

        conf_val = _get_value(conf, child)
        try:
            parsed_val = child.parse_value(conf_val, test_mode)
            parsed_children[child.name] = parsed_val
        except exceptions.InvalidConfValue as e:
            errors[child.path_str()] = e.errors

    if errors:
        raise exceptions.InvalidConfGroup(errors)

    if lazy_loading:
        return parsed_children
    return groups.GroupNamespace(group.name, parsed_children)


def load_config_dict(config_group, conf_dict, root_group_name="conf", test_mode=False):
    """Load the config group

    :param config_group: A single config group.
    :param conf_dict: The loaded dictionary of config values
    :param root_group_name: If config_groups is a list, the namespace of the
                            root wrapper group.
    :param test_mode: If test are running. Will ignore deprecation warnings
                      and ignore fields with 'required_for_tests=False'.
    """
    try:
        parsed_group = process_group(config_group, conf_dict, test_mode)
    except exceptions.InvalidConfGroup as e:
        raise exceptions.InvalidConf({config_group.name: e.errors})

    return parsed_group


def load_config_file(
    config_group,
    conf_file,
    config_format=None,
    root_group_name="conf",
    test_mode=False,
):
    """Load the config group from a file

    :param config_group: A single config group.
    :param conf_file: Config file location.
    :param config_format: The format the config file is in: yaml or toml
    :param root_group_name: If config_groups is a list, the namespace of the
                            root wrapper group.
    :param test_mode: If test are running. Will ignore deprecation warnings
                      and ignore fields with 'required_for_tests=False'.
    """

    if not config_format:
        config_format = utils.validate_config_format(conf_file.split(".")[-1])
    else:
        config_format = utils.validate_config_format(config_format)

    conf_dict = None
    if config_format == "yaml":
        with open(conf_file) as f:
            conf_dict = yaml.load(f, Loader=yaml.FullLoader)
            conf_dict = {config_group.name: conf_dict}
    elif config_format == "toml":
        conf_dict = toml.load(conf_file)

    return load_config_dict(config_group, conf_dict, root_group_name, test_mode)



# Copyright (C) 2019 Catalyst Cloud Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from collections import OrderedDict

from slugify import slugify

from confspirator import base
from confspirator import exceptions
from confspirator.fields import BaseConfigField
from confspirator import utils


class ConfigGroup(base.BaseConfigGroup):
    """Configuration group class

    Group which can onto it have registered fields or nested
    groups.

    :param name: the name of this group
    :param description: a description of the group
    :param lazy_load: if true the group won't fully load until accessed
    :param children: children to register right away
    """

    def __init__(
        self, name, description=None, children=None, lazy_load=False, reformat_name=True
    ):
        if reformat_name:
            name = slugify(name)
        # we still at the very least do some basic formatting
        name = name.replace("-", "_").replace(" ", "_")
        if name.startswith("_"):
            raise ValueError("illegal name %s with prefix _" % (name,))
        self.name = name
        self.description = description
        self.lazy_load = lazy_load
        self._parent_group = None
        self._children = OrderedDict()
        if children:
            for child in children:
                self.register_child_config(child)

    def set_parent(self, parent):
        self._parent_group = parent

    def path(self):
        """Get the path down to this group"""
        path = []
        parent = self._parent_group
        while parent:
            path.append(parent.name)
            parent = parent._parent_group
        path.reverse()
        path.append(self.name)
        return path

    def register_child_config(self, child, override=False):
        """Register a child config

        :param child: a ConfigField or ConfigGroup instance
        :param override: replace if name already present
        """
        if not isinstance(child, (BaseConfigField, ConfigGroup)):
            raise exceptions.InvalidConfigClass(
                "'%s' is not a valid config class" % child
            )
        if not override and child.name in self._children:
            raise exceptions.ConfigGroupAlreadyRegistered(
                "'%s' is already registered in this group." % child
            )
        self._children[child.name] = child
        child.set_parent(self)

    def extend(self, children=None, remove_children=None):
        """Make a new group based on this group

        :param children: a list of configs to override or add
        :param remove_children: A list of names to remove
        """
        new_children = {}
        for name, child in self._children.items():
            if isinstance(child, ConfigGroup):
                new_children[name] = copy.deepcopy(child)
            else:
                new_children[name] = copy.copy(child)
        if remove_children:
            for name in remove_children:
                new_children.pop(name, None)

        new_group = ConfigGroup(
            self.name, children=new_children.values(), lazy_load=self.lazy_load
        )

        if children:
            for child in children:
                new_group.register_child_config(child, override=True)

        return new_group

    def copy(self):
        return copy.deepcopy(self)

    def __contains__(self, key):
        """Return True if key is the name of a registered config or group."""
        return key in self._children

    def __iter__(self):
        """Iterate over all registered config and groups.

        Always start with fields, then groups.
        """
        fields = []
        groups = []
        for child in self._children.values():
            if isinstance(child, ConfigGroup):
                groups.append(child)
            else:
                fields.append(child)
        children = fields + groups
        for child in children:
            yield child

    def __str__(self):
        return self.name

    def __deepcopy__(self, memo):
        new_children = []
        for child in self._children.values():
            if isinstance(child, ConfigGroup):
                new_children.append(copy.deepcopy(child, memo=memo))
            else:
                new_children.append(copy.copy(child))

        new_group = ConfigGroup(
            self.name,
            lazy_load=self.lazy_load,
            children=new_children,
            reformat_name=False,
        )
        new_group.set_parent(self._parent_group)
        return new_group


class DynamicNameConfigGroup(ConfigGroup):
    """Configuration group class which must be named later

    Group which can onto it have registered fields or nested
    groups.

    :param description: a description of the group
    :param lazy_load: if true the group won't fully load until accessed
    :param children: children to register right away
    """

    def __init__(self, description=None, children=None, lazy_load=False):
        self.lazy_load = lazy_load
        self.description = description
        self._parent_group = None
        self._children = OrderedDict()
        if children:
            for child in children:
                self.register_child_config(child)

        self._name = None

    def set_name(self, name, reformat_name=True):
        if reformat_name:
            name = slugify(name)
        # we still at the very least do some basic formatting
        name = name.replace("-", "_").replace(" ", "_")
        if name.startswith("_"):
            raise ValueError("illegal name %s with prefix _" % (name,))
        self._name = name

    @property
    def name(self):
        if not self._name:
            raise exceptions.InvalidConfGroup("Must name the group before using it.")
        return self._name

    def extend(self, children=None, remove_children=None):
        new_children = {}
        for name, child in self._children.items():
            if isinstance(child, ConfigGroup):
                new_children[name] = copy.deepcopy(child)
            else:
                new_children[name] = copy.copy(child)
        if remove_children:
            for name in remove_children:
                new_children.pop(name, None)

        new_group = DynamicNameConfigGroup(
            children=new_children.values(), lazy_load=self.lazy_load
        )

        if children:
            for child in children:
                new_group.register_child_config(child, override=True)

        return new_group

    def __deepcopy__(self, memo):
        new_children = []
        for child in self._children.values():
            if isinstance(child, ConfigGroup):
                new_children.append(copy.deepcopy(child, memo=memo))
            else:
                new_children.append(copy.copy(child))

        new_group = DynamicNameConfigGroup(
            children=new_children,
            lazy_load=self.lazy_load,
        )
        new_group.set_parent(self._parent_group)
        try:
            new_group.set_name(self.name, reformat_name=False)
        except exceptions.InvalidConfGroup:
            pass
        return new_group


class GroupNamespace(base.BaseGroupNamespace):
    """A namespace in which to store parsed config data

    This is a parsed version of a config group.

    :param name: the name of the group
    :param values: a dict of the child values
    """

    def __init__(self, name, values):
        self.name = name
        self._parent = None

        new_values = {}
        for name, child in values.items():
            if isinstance(child, GroupNamespace):
                child.set_parent(self)
                new_values[child.name] = child
            else:
                new_values[name] = child
        self._values = new_values

    def set_parent(self, parent):
        self._parent = parent

    def path(self):
        path = []
        parent = self._parent
        while parent:
            path.append(parent.name)
            parent = parent._parent
        path.reverse()
        path.append(self.name)
        return path

    def overlay(self, overriding_group):
        """Merge a group or a dict to overlay the values in this one

        Returns a copy with the merged contents.

        :param overriding_group: the group or dict which to overlay
        """
        return utils.recursive_merge(self, overriding_group)

    def __getattr__(self, name):
        try:
            return self._values[name]
        except KeyError:
            path = self.path() + [name]
            raise exceptions.NoSuchConfig(".".join(path))

    def __getitem__(self, key):
        return self.__getattr__(key)

    def get(self, key):
        return self.__getattr__(key)

    def __contains__(self, key):
        return key in self._values

    def __str__(self):
        return "ConfigNameSpace: " + ".".join(self.path())

    def __deepcopy__(self, memo):
        new_values = {}
        for k, v in self._values.items():
            new_values[k] = copy.deepcopy(v, memo=memo)

        new_group = GroupNamespace(self.name, new_values)
        new_group.set_parent(self._parent)
        return new_group


class LazyLoadedGroupNamespace(GroupNamespace):
    """A lazy loaded groupnamespace

    This version will not process the group it is for until
    the group is accessed at least once. This allows for extra
    configs to be added to the group at runtime or app startup
    and for the conf for this group only to be parsed when that
    registration is well and truly finished.

    :param name: the name of the group
    :param group: the group which we want to lazy load later
    :param conf: the config dictionary we want to parse
    :param test_mode: are we running in test mode
    """

    def __init__(self, name, group, conf, test_mode):
        self.name = name
        self._group = group
        self._conf = conf
        self._test_mode = test_mode

        self._values = None
        self._parent = None

    def _lazy_load(self):
        if self._values is None:
            from confspirator import loader

            self._values = loader.process_group(
                self._group, self._conf, self._test_mode, lazy_loading=True
            )

            delattr(self, "_group")
            delattr(self, "_conf")
            delattr(self, "_test_mode")

            for child in self._values.values():
                if isinstance(child, GroupNamespace):
                    child.set_parent(self)

    def __getattr__(self, name):
        self._lazy_load()
        return super(LazyLoadedGroupNamespace, self).__getattr__(name)

    def __deepcopy__(self, memo):
        self._lazy_load()
        return super(LazyLoadedGroupNamespace, self).__deepcopy__(memo)



# Copyright (C) 2019 Catalyst Cloud Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import unittest
from mock import patch


import confspirator
from confspirator import exceptions
from confspirator import groups
from confspirator import fields
from confspirator import types


class ConfigTests(unittest.TestCase):
    def test_config_register_same(self):
        config_group = groups.ConfigGroup("test")
        config_group.register_child_config(fields.StrConfig("some_value"))
        with self.assertRaises(exceptions.ConfigGroupAlreadyRegistered):
            config_group.register_child_config(fields.StrConfig("some_value"))

    def test_str_config_field_getters(self):
        default_val = "This is a default"
        config_group = groups.ConfigGroup("test")
        config_group.register_child_config(
            fields.StrConfig("some_value", default=default_val)
        )

        config = confspirator.load(config_group, {})
        self.assertEqual(config.get("some_value"), default_val)
        self.assertEqual(config["some_value"], default_val)
        self.assertEqual(config.some_value, default_val)

    def test_str_config_field_no_conf(self):
        config_group = groups.ConfigGroup("test")
        config_group.register_child_config(fields.StrConfig("some_value"))

        config = confspirator.load(config_group, {})
        with self.assertRaises(exceptions.NoSuchConfig):
            config.get("no_value")

    def test_str_config_field_with_default(self):
        default_val = "This is a default"

        config_group = groups.ConfigGroup("test")
        config_group.register_child_config(
            fields.StrConfig("some_value", required=True, default=default_val)
        )

        config = confspirator.load(config_group, {})
        self.assertEqual(config.some_value, default_val)

    def test_str_config_field_with_required(self):
        config_group = groups.ConfigGroup("test")
        config_group.register_child_config(
            fields.StrConfig("some_value", required=True)
        )

        with self.assertRaises(exceptions.InvalidConf):
            confspirator.load(config_group, {})

    def test_str_config_field_no_default_not_required(self):
        config_group = groups.ConfigGroup("test")
        config_group.register_child_config(fields.StrConfig("some_value"))

        config = confspirator.load(config_group, {})
        self.assertEqual(config.some_value, None)

    def test_str_config_field_with_required_for_tests(self):
        config_group = groups.ConfigGroup("test")
        config_group.register_child_config(
            fields.StrConfig("some_value", required=True, required_for_tests=False)
        )

        confspirator.load(config_group, {}, test_mode=True)

    def test_str_config_field_with_default_for_tests(self):
        config_group = groups.ConfigGroup("test")
        default_val = "other value"
        test_val = "some value"
        config_group.register_child_config(
            fields.StrConfig("some_value", default=default_val, test_default=test_val)
        )

        config = confspirator.load(config_group, {}, test_mode=True)
        self.assertEqual(config.some_value, test_val)
        config = confspirator.load(config_group, {})
        self.assertEqual(config.some_value, default_val)

    def test_str_config_nested(self):
        config_group = groups.ConfigGroup("test")
        config_group2 = groups.ConfigGroup("test2")
        config_group.register_child_config(config_group2)
        config_group3 = groups.ConfigGroup("test3")
        config_group2.register_child_config(config_group3)
        config_group3.register_child_config(
            fields.StrConfig("some_value", required=True)
        )

        val = "This is a value"
        conf_dict = {"test": {"test2": {"test3": {"some_value": val}}}}
        config = confspirator.load(config_group, conf_dict)
        self.assertEqual(config.test2.test3.some_value, val)

    def test_str_config_envvar(self):
        config_group = groups.ConfigGroup("test")
        config_group2 = groups.ConfigGroup("test2")
        config_group.register_child_config(config_group2)
        config_group2.register_child_config(
            fields.StrConfig("some_value", required=True)
        )

        val = "This is a value"
        with patch.dict("os.environ", {"TEST_TEST2_SOME_VALUE": val}):
            config = confspirator.load(config_group, {})
        self.assertEqual(config.test2.some_value, val)

    def test_str_config_deprecated_location(self):
        config_group = groups.ConfigGroup("test")
        config_group2 = groups.ConfigGroup("test2")
        config_group.register_child_config(config_group2)
        config_group2.register_child_config(
            fields.StrConfig(
                "some_value",
                required=True,
                deprecated_location="test.test2.test3.old_value",
            )
        )

        val = "This is a value"
        conf_dict = {"test": {"test2": {"test3": {"old_value": val}}}}
        config = confspirator.load(config_group, conf_dict)
        self.assertEqual(config.test2.some_value, val)

    def test_bool_config_field_with_no_default_not_required(self):
        """no default, and not required will return a None"""
        config_group = groups.ConfigGroup("test")
        config_group.register_child_config(fields.BoolConfig("some_value"))

        config = confspirator.load(config_group, {})
        self.assertEqual(config.some_value, None)

    def test_config_group_lazy_load(self):
        config_group = groups.ConfigGroup("test", lazy_load=True)
        config_group.register_child_config(
            fields.StrConfig("some_value", required=True)
        )

        val = "This is a value"
        conf_dict = {"test": {"some_value": val}}
        config = confspirator.load(config_group, conf_dict)
        self.assertEqual(config._values, None)
        self.assertEqual(config._conf, conf_dict)
        self.assertEqual(config.some_value, val)
        self.assertEqual(config._values, {"some_value": val})

    def test_config_overlay_basic(self):
        config_group = groups.ConfigGroup("test")
        config_group2 = groups.ConfigGroup("test2")
        config_group.register_child_config(config_group2)
        config_group3 = groups.ConfigGroup("test3")
        config_group2.register_child_config(config_group3)
        config_group3.register_child_config(
            fields.StrConfig("some_value", required=True)
        )

        val = "This is a value"
        conf_dict = {"test": {"test2": {"test3": {"some_value": val}}}}
        config = confspirator.load(config_group, conf_dict)
        self.assertEqual(config.test2.test3.some_value, val)

        val = "This is another value"
        conf_dict = {"test2": {"test3": {"some_value": val}}}
        new_conf = config.overlay(conf_dict)
        self.assertEqual(new_conf.test2.test3.some_value, val)

    def test_config_overlay_group(self):
        config_group = groups.ConfigGroup("test")
        config_group2 = groups.ConfigGroup("test2")
        config_group2.register_child_config(fields.StrConfig("a_value", required=True))
        config_group.register_child_config(config_group2)
        config_group25 = groups.ConfigGroup("test25")
        config_group.register_child_config(config_group25)
        config_group25.register_child_config(
            fields.StrConfig("some_value", required=True)
        )

        val = "This is a value"
        val2 = "This is another value"
        conf_dict = {
            "test": {"test2": {"a_value": val}, "test25": {"some_value": val2}}
        }
        config = confspirator.load(config_group, conf_dict)
        self.assertEqual(config.test2.a_value, val)
        self.assertEqual(config.test25.some_value, val2)

        new_conf = config.test2.overlay(config.test25)
        self.assertEqual(new_conf.some_value, val2)

    def test_config_overlay_complex(self):
        config_group = groups.ConfigGroup("test")
        config_group2 = groups.ConfigGroup("test2")
        config_group.register_child_config(config_group2)
        config_group3 = groups.ConfigGroup("test3")
        config_group35 = groups.ConfigGroup("test35")
        config_group2.register_child_config(config_group3)
        config_group2.register_child_config(config_group35)
        config_group3.register_child_config(
            fields.StrConfig("some_value", required=True)
        )
        config_group35.register_child_config(
            fields.DictConfig("some_dict", required=True)
        )

        val = "This is a value"
        val2 = "This is a value2"
        conf_dict = {
            "test": {
                "test2": {
                    "test3": {"some_value": val},
                    "test35": {"some_dict": {"some_inner_val": val2}},
                }
            }
        }
        config = confspirator.load(config_group, conf_dict)
        self.assertEqual(config.test2.test3.some_value, val)
        self.assertEqual(config.test2.test35.some_dict["some_inner_val"], val2)

        val = "This is another value"
        val2 = "This is another value2"
        conf_dict = {
            "test2": {
                "test3": {"some_value": val},
                "test35": {"some_dict": {"some_inner_val": val2}},
            }
        }
        new_conf = config.overlay(conf_dict)
        self.assertEqual(new_conf.test2.test3.some_value, val)
        self.assertEqual(new_conf.test2.test35.some_dict["some_inner_val"], val2)

    def test_config_overlay_override_with_none(self):
        config_group = groups.ConfigGroup("test")
        config_group2 = groups.ConfigGroup("test2")
        config_group.register_child_config(config_group2)
        config_group3 = groups.ConfigGroup("test3")
        config_group2.register_child_config(config_group3)
        config_group3.register_child_config(
            fields.StrConfig("some_value", required=True)
        )

        val = "This is a value"
        conf_dict = {"test": {"test2": {"test3": {"some_value": val}}}}
        config = confspirator.load(config_group, conf_dict)
        self.assertEqual(config.test2.test3.some_value, val)

        val = "This is another value"
        conf_dict = {"test2": {"test3": None}}
        new_conf = config.overlay(conf_dict)
        self.assertEqual(new_conf.test2.test3, None)

    def test_config_group_extend(self):
        config_group = groups.ConfigGroup("test")
        config_group2 = groups.ConfigGroup("test2")
        config_group2.register_child_config(
            fields.StrConfig("some_value", required=True)
        )
        config_group3 = groups.ConfigGroup("test3")
        config_group2.register_child_config(config_group3)
        config_group3.register_child_config(
            fields.StrConfig("some_deeper_value", required=True)
        )
        extended_group2 = config_group2.extend(
            children=[fields.StrConfig("some_other_value", required=True)]
        )
        config_group.register_child_config(extended_group2)

        val = "This is a value"
        conf_dict = {"test": {"test2": {"some_value": val}}}
        with self.assertRaises(exceptions.InvalidConf):
            config = confspirator.load(config_group, conf_dict)

        val2 = "This is also a value"
        val3 = "This is also a value"
        conf_dict = {
            "test": {
                "test2": {
                    "some_value": val,
                    "some_other_value": val2,
                    "test3": {"some_deeper_value": val3},
                }
            }
        }
        config = confspirator.load(config_group, conf_dict)
        self.assertEqual(config.test2.some_value, val)
        self.assertEqual(config.test2.some_other_value, val2)
        self.assertEqual(config.test2.test3.some_deeper_value, val2)

    def test_config_group_extend_dynamic(self):
        config_group = groups.ConfigGroup("test")
        config_group2 = groups.DynamicNameConfigGroup()
        config_group2.register_child_config(
            fields.StrConfig("some_value", required=True)
        )
        config_group25 = config_group2.extend(
            children=[fields.StrConfig("some_other_value", required=True)],
            remove_children=["some_value"],
        )
        config_group2.set_name("test2")
        config_group25.set_name("test25")
        config_group.register_child_config(config_group2)
        config_group.register_child_config(config_group25)

        val = "This is a value"
        conf_dict = {"test": {"test2": {"some_value": val}}}
        with self.assertRaises(exceptions.InvalidConf):
            config = confspirator.load(config_group, conf_dict)

        val2 = "This is also a value"
        conf_dict = {
            "test": {"test2": {"some_value": val}, "test25": {"some_other_value": val2}}
        }
        config = confspirator.load(config_group, conf_dict)
        self.assertEqual(config.test2.some_value, val)
        self.assertEqual(config.test25.some_other_value, val2)

    def test_config_group_deep_copy(self):
        config_group = groups.ConfigGroup("test")
        config_group2 = groups.ConfigGroup("test2")
        config_group.register_child_config(config_group2)
        config_group2.register_child_config(
            fields.StrConfig("some_value", required=True)
        )

        new_group = config_group.copy()
        new_group._children["test2"].register_child_config(
            fields.StrConfig("some_other_value", required=True)
        )

        val = "This is a value"
        conf_dict = {"test": {"test2": {"some_value": val}}}
        config = confspirator.load(config_group, conf_dict)
        self.assertEqual(config.test2.some_value, val)

        with self.assertRaises(exceptions.InvalidConf):
            config = confspirator.load(new_group, conf_dict)

        val2 = "This is also a value"
        conf_dict = {"test": {"test2": {"some_value": val, "some_other_value": val2}}}
        config = confspirator.load(new_group, conf_dict)
        self.assertEqual(config.test2.some_other_value, val2)

    def test_config_group_deep_copy_and_rename(self):
        config_group = groups.ConfigGroup("test")
        config_group2 = groups.DynamicNameConfigGroup()
        config_group2.register_child_config(
            fields.StrConfig("some_value", required=True)
        )
        config_group25 = config_group2.copy()
        config_group2.set_name("test2")
        config_group25.set_name("test25")
        config_group.register_child_config(config_group2)
        config_group.register_child_config(config_group25)

        val = "This is a value"
        conf_dict = {"test": {"test2": {"some_value": val}}}
        with self.assertRaises(exceptions.InvalidConf):
            config = confspirator.load(config_group, conf_dict)

        val2 = "This is also a value"
        conf_dict = {
            "test": {"test2": {"some_value": val}, "test25": {"some_value": val2}}
        }
        config = confspirator.load(config_group, conf_dict)
        self.assertEqual(config.test2.some_value, val)
        self.assertEqual(config.test25.some_value, val2)


class ConfigTypeTests(unittest.TestCase):
    def test_str_config_with_length(self):
        config_group = groups.ConfigGroup("test")
        config_group.register_child_config(fields.StrConfig("some_value", max_length=6))

        val = "1234567"
        conf_dict = {"test": {"some_value": val}}
        with self.assertRaises(exceptions.InvalidConf):
            confspirator.load(config_group, conf_dict)
        val = "123456"
        conf_dict = {"test": {"some_value": val}}
        config = confspirator.load(config_group, conf_dict)
        self.assertEqual(config.some_value, val)

    def test_str_config_with_regex(self):
        config_group = groups.ConfigGroup("test")
        config_group.register_child_config(
            fields.StrConfig(
                "some_value", regex=r"#?([\da-fA-F]{2})([\da-fA-F]{2})([\da-fA-F]{2})"
            )
        )

        val = "not_hex"
        conf_dict = {"test": {"some_value": val}}
        with self.assertRaises(exceptions.InvalidConf):
            confspirator.load(config_group, conf_dict)
        val = "#F0F0F0"
        conf_dict = {"test": {"some_value": val}}
        config = confspirator.load(config_group, conf_dict)
        self.assertEqual(config.some_value, val)

    def test_str_config_with_choices(self):
        config_group = groups.ConfigGroup("test")
        config_group.register_child_config(
            fields.StrConfig("some_value", choices=["opt1", "opt2"])
        )

        val = "opt3"
        conf_dict = {"test": {"some_value": val}}
        with self.assertRaises(exceptions.InvalidConf):
            confspirator.load(config_group, conf_dict)
        val = "opt2"
        conf_dict = {"test": {"some_value": val}}
        config = confspirator.load(config_group, conf_dict)
        self.assertEqual(config.some_value, val)

    def test_bool_config(self):
        config_group = groups.ConfigGroup("test")
        config_group.register_child_config(fields.BoolConfig("some_value"))

        val = "nada"
        conf_dict = {"test": {"some_value": val}}
        with self.assertRaises(exceptions.InvalidConf):
            confspirator.load(config_group, conf_dict)
        val = False
        conf_dict = {"test": {"some_value": val}}
        config = confspirator.load(config_group, conf_dict)
        self.assertEqual(config.some_value, val)
        val = True
        conf_dict = {"test": {"some_value": val}}
        config = confspirator.load(config_group, conf_dict)
        self.assertEqual(config.some_value, val)
        val = "false"
        conf_dict = {"test": {"some_value": val}}
        config = confspirator.load(config_group, conf_dict)
        self.assertEqual(config.some_value, False)
        val = "true"
        conf_dict = {"test": {"some_value": val}}
        config = confspirator.load(config_group, conf_dict)
        self.assertEqual(config.some_value, True)

    def test_int_config(self):
        config_group = groups.ConfigGroup("test")
        config_group.register_child_config(fields.IntConfig("some_value"))

        val = "nada"
        conf_dict = {"test": {"some_value": val}}
        with self.assertRaises(exceptions.InvalidConf):
            confspirator.load(config_group, conf_dict)
        val = "5"
        conf_dict = {"test": {"some_value": val}}
        config = confspirator.load(config_group, conf_dict)
        self.assertEqual(config.some_value, int(val))
        val = 5
        conf_dict = {"test": {"some_value": val}}
        config = confspirator.load(config_group, conf_dict)
        self.assertEqual(config.some_value, val)

    def test_int_config_min_max(self):
        config_group = groups.ConfigGroup("test")
        config_group.register_child_config(
            fields.IntConfig("some_value", min=5, max=10)
        )

        val = 4
        conf_dict = {"test": {"some_value": val}}
        with self.assertRaises(exceptions.InvalidConf):
            confspirator.load(config_group, conf_dict)
        val = 11
        conf_dict = {"test": {"some_value": val}}
        with self.assertRaises(exceptions.InvalidConf):
            confspirator.load(config_group, conf_dict)
        val = 6
        conf_dict = {"test": {"some_value": val}}
        config = confspirator.load(config_group, conf_dict)
        self.assertEqual(config.some_value, val)

    def test_float_config(self):
        config_group = groups.ConfigGroup("test")
        config_group.register_child_config(fields.FloatConfig("some_value"))

        val = "nada"
        conf_dict = {"test": {"some_value": val}}
        with self.assertRaises(exceptions.InvalidConf):
            confspirator.load(config_group, conf_dict)
        val = "5.2"
        conf_dict = {"test": {"some_value": val}}
        config = confspirator.load(config_group, conf_dict)
        self.assertEqual(config.some_value, float(val))
        val = 5.5
        conf_dict = {"test": {"some_value": val}}
        config = confspirator.load(config_group, conf_dict)
        self.assertEqual(config.some_value, val)

    def test_list_config_from_list(self):
        config_group = groups.ConfigGroup("test")
        config_group.register_child_config(
            fields.ListConfig("some_list", required=True)
        )

        val = ["this", "is", "a", "comma", "separated", "list"]
        conf_dict = {"test": {"some_list": val}}
        config = confspirator.load(config_group, conf_dict)
        self.assertEqual(config.some_list, val)

    def test_list_config_from_str(self):
        config_group = groups.ConfigGroup("test")
        config_group.register_child_config(
            fields.ListConfig("some_list", required=True)
        )

        val = "this,is,a,comma,separated,list"
        conf_dict = {"test": {"some_list": val}}
        config = confspirator.load(config_group, conf_dict)
        self.assertEqual(
            config.some_list, ["this", "is", "a", "comma", "separated", "list"]
        )

    def test_list_config_from_json_dicts(self):
        config_group = groups.ConfigGroup("test")
        config_group.register_child_config(
            fields.ListConfig(
                "some_list", item_type=types.Dict(is_json=True), required=True
            )
        )

        val = [
            '{"key1": {"key2": 5}}',
            '{"key1": 7}',
        ]
        conf_dict = {"test": {"some_list": val}}
        config = confspirator.load(config_group, conf_dict)
        self.assertEqual(config.some_list, [{"key1": {"key2": 5}}, {"key1": 7}])

    def test_dict_config_from_dict(self):
        config_group = groups.ConfigGroup("test")
        config_group.register_child_config(
            fields.DictConfig("some_dict", required=True)
        )

        val = {"dict": "config"}
        conf_dict = {"test": {"some_dict": val}}
        config = confspirator.load(config_group, conf_dict)
        self.assertEqual(config.some_dict, val)

    def test_dict_config_from_str(self):
        config_group = groups.ConfigGroup("test")
        config_group.register_child_config(
            fields.DictConfig("some_dict", required=True)
        )

        val = "dict:config,dict2:config2"
        conf_dict = {"test": {"some_dict": val}}
        config = confspirator.load(config_group, conf_dict)
        self.assertEqual(config.some_dict, {"dict": "config", "dict2": "config2"})

    def test_dict_config_from_str_as_json(self):
        config_group = groups.ConfigGroup("test")
        config_group.register_child_config(
            fields.DictConfig("some_dict", required=True, is_json=True)
        )

        val = '{"dict": "config", "dict2": "config2"}'
        conf_dict = {"test": {"some_dict": val}}
        config = confspirator.load(config_group, conf_dict)
        self.assertEqual(config.some_dict, {"dict": "config", "dict2": "config2"})

    def test_dict_config_from_dict_nested(self):
        config_group = groups.ConfigGroup("test")
        config_group.register_child_config(
            fields.DictConfig(
                "some_dict",
                required=True,
                value_type=types.Dict(),
                check_value_type=True,
            )
        )

        val = {"dict": "config"}
        conf_dict = {"test": {"some_dict": val}}
        with self.assertRaises(exceptions.InvalidConf):
            confspirator.load(config_group, conf_dict)
        val = {"dict": {"thing": "config"}}
        conf_dict = {"test": {"some_dict": val}}
        config = confspirator.load(config_group, conf_dict)
        self.assertEqual(config.some_dict, val)

    def test_dict_config_from_json_nested(self):
        config_group = groups.ConfigGroup("test")
        config_group.register_child_config(
            fields.DictConfig(
                "some_dict",
                required=True,
                value_type=types.Dict(),
                check_value_type=True,
                is_json=True,
            )
        )

        val = '{"dict": "config"}'
        conf_dict = {"test": {"some_dict": val}}
        with self.assertRaises(exceptions.InvalidConf):
            confspirator.load(config_group, conf_dict)
        val = '{"dict": {"thing": "config"}}'
        conf_dict = {"test": {"some_dict": val}}
        config = confspirator.load(config_group, conf_dict)
        self.assertEqual(config.some_dict, {"dict": {"thing": "config"}})

    def test_ip_config(self):
        config_group = groups.ConfigGroup("test")
        config_group.register_child_config(fields.IPConfig("some_value"))

        val = "not_an_ip"
        conf_dict = {"test": {"some_value": val}}
        with self.assertRaises(exceptions.InvalidConf):
            confspirator.load(config_group, conf_dict)
        val = "0.0.0.0"
        conf_dict = {"test": {"some_value": val}}
        config = confspirator.load(config_group, conf_dict)
        self.assertEqual(config.some_value, val)
        val = "2001:db8::8a2e:370:7334"
        conf_dict = {"test": {"some_value": val}}
        config = confspirator.load(config_group, conf_dict)
        self.assertEqual(config.some_value, val)
        val = "2001:0db8:0000:0000:0000:8a2e:0370:7334"
        conf_dict = {"test": {"some_value": val}}
        config = confspirator.load(config_group, conf_dict)
        self.assertEqual(config.some_value, val)

    def test_ip_config_version(self):
        config_group = groups.ConfigGroup("test")
        config_group.register_child_config(fields.IPConfig("some_value", version=4))

        val = "2001:db8::8a2e:370:7334"
        conf_dict = {"test": {"some_value": val}}
        with self.assertRaises(exceptions.InvalidConf):
            confspirator.load(config_group, conf_dict)
        val = "0.0.0.0"
        conf_dict = {"test": {"some_value": val}}
        config = confspirator.load(config_group, conf_dict)
        self.assertEqual(config.some_value, val)

        config_group = groups.ConfigGroup("test")
        config_group.register_child_config(fields.IPConfig("some_value", version=6))
        val = "0.0.0.0"
        conf_dict = {"test": {"some_value": val}}
        with self.assertRaises(exceptions.InvalidConf):
            confspirator.load(config_group, conf_dict)
        val = "2001:db8::8a2e:370:7334"
        conf_dict = {"test": {"some_value": val}}
        config = confspirator.load(config_group, conf_dict)
        self.assertEqual(config.some_value, val)

    def test_port_config(self):
        config_group = groups.ConfigGroup("test")
        config_group.register_child_config(fields.PortConfig("some_value"))

        val = "9999999"
        conf_dict = {"test": {"some_value": val}}
        with self.assertRaises(exceptions.InvalidConf):
            confspirator.load(config_group, conf_dict)
        val = "8888"
        conf_dict = {"test": {"some_value": val}}
        config = confspirator.load(config_group, conf_dict)
        self.assertEqual(config.some_value, int(val))
        val = 8888
        conf_dict = {"test": {"some_value": val}}
        config = confspirator.load(config_group, conf_dict)
        self.assertEqual(config.some_value, val)

    def test_hostname_config(self):
        config_group = groups.ConfigGroup("test")
        config_group.register_child_config(fields.HostNameConfig("some_value"))

        val = "this#is.not_a_valid*hostname"
        conf_dict = {"test": {"some_value": val}}
        with self.assertRaises(exceptions.InvalidConf):
            confspirator.load(config_group, conf_dict)
        val = "this.is.a.valid.hostname"
        conf_dict = {"test": {"some_value": val}}
        config = confspirator.load(config_group, conf_dict)
        self.assertEqual(config.some_value, val)

    def test_hostaddress_config(self):
        config_group = groups.ConfigGroup("test")
        config_group.register_child_config(fields.HostAddressConfig("some_value"))

        val = "this#is.not_a_valid*hostname"
        conf_dict = {"test": {"some_value": val}}
        with self.assertRaises(exceptions.InvalidConf):
            confspirator.load(config_group, conf_dict)
        val = "this.is.a.valid.hostname"
        conf_dict = {"test": {"some_value": val}}
        config = confspirator.load(config_group, conf_dict)
        self.assertEqual(config.some_value, val)
        val = "0.0.0.0"
        conf_dict = {"test": {"some_value": val}}
        config = confspirator.load(config_group, conf_dict)
        self.assertEqual(config.some_value, val)

    def test_uri_config(self):
        config_group = groups.ConfigGroup("test")
        config_group.register_child_config(fields.URIConfig("some_value"))

        val = "this#is.not_a_valid*hostname"
        conf_dict = {"test": {"some_value": val}}
        with self.assertRaises(exceptions.InvalidConf):
            confspirator.load(config_group, conf_dict)
        val = "https://this.is.a.valid.hostname"
        conf_dict = {"test": {"some_value": val}}
        config = confspirator.load(config_group, conf_dict)
        self.assertEqual(config.some_value, val)
        val = "ftp://this.is.a.valid.hostname"
        conf_dict = {"test": {"some_value": val}}
        config = confspirator.load(config_group, conf_dict)
        self.assertEqual(config.some_value, val)

    def test_uri_config_scheme(self):
        config_group = groups.ConfigGroup("test")
        config_group.register_child_config(
            fields.URIConfig("some_value", schemes=["https", "http"])
        )

        val = "ftp://this.is.a.valid.hostname"
        conf_dict = {"test": {"some_value": val}}
        with self.assertRaises(exceptions.InvalidConf):
            confspirator.load(config_group, conf_dict)
        val = "https://this.is.a.valid.hostname"
        conf_dict = {"test": {"some_value": val}}
        config = confspirator.load(config_group, conf_dict)
        self.assertEqual(config.some_value, val)
        val = "http://this.is.a.valid.hostname"
        conf_dict = {"test": {"some_value": val}}
        config = confspirator.load(config_group, conf_dict)
        self.assertEqual(config.some_value, val)



# Copyright (C) 2020 Catalyst Cloud Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import unittest


from confspirator import example
from confspirator import groups
from confspirator import fields


class ExampleTests(unittest.TestCase):

    maxDiff = None

    def _simple_nested_config(self):
        config_group = groups.ConfigGroup("test")
        config_group2 = groups.ConfigGroup("test2")
        config_group.register_child_config(config_group2)
        config_group3 = groups.ConfigGroup("test3")
        config_group2.register_child_config(config_group3)
        config_group3.register_child_config(
            fields.StrConfig("some_value", required=True)
        )
        return config_group

    def _complex_nested_config(self):
        config_group = groups.ConfigGroup("test")
        config_group2 = groups.ConfigGroup("test2")
        config_group.register_child_config(config_group2)
        config_group3 = groups.ConfigGroup("test3")
        config_group2.register_child_config(config_group3)
        config_group3.register_child_config(
            fields.StrConfig(
                "some_value",
                help_text="help text for some_value",
                required=True,
                default="stuff",
            )
        )

        config_group2.register_child_config(
            fields.DictConfig(
                "some_dict",
                help_text="a dict field with a large default.",
                default={
                    "level1": {
                        "level2": {
                            "level3": {"level4": {"some_bool": False, "some_int": 5}}
                        }
                    },
                    "level1a": {"level2a": {"level3a": {"stuff": "things"}}},
                },
            )
        )
        config_group2.register_child_config(
            fields.DictConfig(
                "some_dict_2",
                help_text="a dict to test a weird circular export edge case",
                default={"l1": {"l2": {"l3": {"l4": {}}}}},
            )
        )
        return config_group

    def _all_fields_config(self):
        config_group = groups.ConfigGroup("test")
        config_group2 = groups.ConfigGroup("test2")
        config_group.register_child_config(config_group2)
        config_group2.register_child_config(
            fields.StrConfig(
                "some_string",
                help_text="help text for some_string",
                default="stuff",
            )
        )
        config_group2.register_child_config(
            fields.IntConfig(
                "some_int",
                default=991,
            )
        )
        config_group2.register_child_config(
            fields.FloatConfig(
                "some_float",
                default=9.81,
            )
        )
        config_group2.register_child_config(
            fields.BoolConfig(
                "is_boolean",
                help_text="boolean field help text",
                default=True,
            )
        )
        config_group2.register_child_config(
            fields.ListConfig(
                "some_list",
                help_text="List things.",
                default=["thing1", "thing2", "thing3"],
            )
        )
        config_group2.register_child_config(
            fields.DictConfig(
                "some_dict",
                help_text="a dict field help text",
                default={"level1": {"level2": {"level3": {"level4": {}}}}},
            )
        )
        config_group2.register_child_config(
            fields.IPConfig("some_ip", default="0.0.0.1")
        )
        config_group2.register_child_config(
            fields.PortConfig("some_port", default="7728")
        )
        config_group2.register_child_config(
            fields.HostNameConfig("some_host", default="place.example.org")
        )
        config_group2.register_child_config(
            fields.HostAddressConfig("some_hostaddress", default="place.example.org")
        )
        config_group2.register_child_config(
            fields.URIConfig("some_uri", default="https://place.example.org")
        )
        return config_group

    def _top_level_config(self):
        """Config fields at the top level"""
        config_group = groups.ConfigGroup("test")
        config_group2 = groups.ConfigGroup("test2")
        config_group.register_child_config(config_group2)
        config_group.register_child_config(
            fields.StrConfig("some_value", default="test")
        )
        config_group2.register_child_config(
            fields.StrConfig("some_value2", default="test2")
        )
        return config_group

    def test_example_config_simple_nested_yaml(self):
        config_group = self._simple_nested_config()

        expected_lines = [
            "test2:",
            "  test3:",
            "    # String",
            "    # some_value: <your_value>",
            "",
        ]
        config_lines = example._create_example_lines(config_group, "yaml")
        self.assertEqual(config_lines, expected_lines)

    def test_example_config_simple_nested_toml(self):
        config_group = self._simple_nested_config()

        expected_lines = [
            "[test.test2]",
            "  [test.test2.test3]",
            "    # String",
            "    # some_value = <your_value>",
            "",
        ]
        config_lines = example._create_example_lines(config_group, "toml")
        self.assertEqual(config_lines, expected_lines)

    def test_example_config_complex_nested_yaml(self):
        config_group = self._complex_nested_config()

        expected_lines = [
            "test2:",
            "  # Dict - a dict field with a large default.",
            "  some_dict:",
            "    level1:",
            "      level2:",
            "        level3:",
            "          level4:",
            "            some_bool: false",
            "            some_int: 5",
            "    level1a:",
            "      level2a:",
            "        level3a:",
            "          stuff: things",
            "  # Dict - a dict to test a weird circular export edge case",
            "  some_dict_2:",
            "    l1:",
            "      l2:",
            "        l3:",
            "          l4: {}",
            "  test3:",
            "    # String - help text for some_value",
            "    some_value: stuff",
            "",
        ]
        config_lines = example._create_example_lines(config_group, "yaml")
        self.assertEqual(config_lines, expected_lines)

    def test_example_config_complex_nested_toml(self):
        config_group = self._complex_nested_config()

        expected_lines = [
            "[test.test2]",
            "  # Dict - a dict field with a large default.",
            "  [test.test2.some_dict]",
            "    [test.test2.some_dict.level1a.level2a.level3a]",
            '    stuff = "things"',
            "    [test.test2.some_dict.level1.level2.level3.level4]",
            "    some_bool = false",
            "    some_int = 5",
            "  # Dict - a dict to test a weird circular export edge case",
            "  [test.test2.some_dict_2]",
            "    [test.test2.some_dict_2.l1.l2.l3.l4]",
            "  [test.test2.test3]",
            "    # String - help text for some_value",
            '    some_value = "stuff"',
            "",
        ]
        config_lines = example._create_example_lines(config_group, "toml")
        self.assertEqual(config_lines, expected_lines)

    def test_example_config_all_fields_yaml(self):
        config_group = self._all_fields_config()

        expected_lines = [
            "test2:",
            "  # String - help text for some_string",
            "  some_string: stuff",
            "  # Integer",
            "  some_int: 991",
            "  # Float",
            "  some_float: 9.81",
            "  # Boolean - boolean field help text",
            "  is_boolean: true",
            "  # List - List things.",
            "  some_list:",
            "    - thing1",
            "    - thing2",
            "    - thing3",
            "  # Dict - a dict field help text",
            "  some_dict:",
            "    level1:",
            "      level2:",
            "        level3:",
            "          level4: {}",
            "  # IPAddress",
            "  some_ip: 0.0.0.1",
            "  # Port",
            "  some_port: '7728'",
            "  # Hostname",
            "  some_host: place.example.org",
            "  # HostAddress",
            "  some_hostaddress: place.example.org",
            "  # URI",
            "  some_uri: https://place.example.org",
            "",
        ]
        config_lines = example._create_example_lines(config_group, "yaml")
        self.assertEqual(config_lines, expected_lines)

    def test_example_config_all_fields_toml(self):
        config_group = self._all_fields_config()

        expected_lines = [
            "[test.test2]",
            "  # String - help text for some_string",
            '  some_string = "stuff"',
            "  # Integer",
            "  some_int = 991",
            "  # Float",
            "  some_float = 9.81",
            "  # Boolean - boolean field help text",
            "  is_boolean = true",
            "  # List - List things.",
            '  some_list = [ "thing1", "thing2", "thing3",]',
            "  # Dict - a dict field help text",
            "  [test.test2.some_dict]",
            "    [test.test2.some_dict.level1.level2.level3.level4]",
            "  # IPAddress",
            '  some_ip = "0.0.0.1"',
            "  # Port",
            '  some_port = "7728"',
            "  # Hostname",
            '  some_host = "place.example.org"',
            "  # HostAddress",
            '  some_hostaddress = "place.example.org"',
            "  # URI",
            '  some_uri = "https://place.example.org"',
            "",
        ]
        self.maxDiff = None
        config_lines = example._create_example_lines(config_group, "toml")
        self.assertEqual(config_lines, expected_lines)

    def test_example_config_top_level_fields_yaml(self):
        config_group = self._top_level_config()

        expected_lines = [
            "# String",
            "some_value: test",
            "",
            "test2:",
            "  # String",
            "  some_value2: test2",
            "",
        ]

        config_lines = example._create_example_lines(config_group, "yaml")
        self.assertEqual(config_lines, expected_lines)

    def test_example_config_top_level_fields_toml(self):
        config_group = self._top_level_config()

        expected_lines = [
            "[test]",
            "# String",
            'some_value = "test"',
            "",
            "[test.test2]",
            "  # String",
            '  some_value2 = "test2"',
            "",
        ]
        config_lines = example._create_example_lines(config_group, "toml")
        self.assertEqual(config_lines, expected_lines)

    def test_example_config_toml_field_ordering(self):
        """Test that output format always places fields before groups

        Fields in a given group need to be placed before any subgroups
        else toml will read them as included in the next group.

        Yaml this isn't an issue since indentation solves it, but for toml
        we need to ensure fields as written down before a subgroup.
        """
        config_group = groups.ConfigGroup("test")
        config_group2 = groups.ConfigGroup("test2")
        config_group.register_child_config(config_group2)
        config_group2a = groups.ConfigGroup("test2a")
        config_group.register_child_config(config_group2a)
        config_group3 = groups.ConfigGroup("test3")
        config_group2.register_child_config(config_group3)
        config_group.register_child_config(
            fields.StrConfig("some_value", required=True)
        )
        config_group.register_child_config(fields.IntConfig("some_int", required=True))
        config_group2.register_child_config(
            fields.StrConfig("some_value2", required=True)
        )
        config_group2a.register_child_config(
            fields.StrConfig("some_value2a", required=True)
        )

        expected_lines = [
            "[test]",
            "# String",
            "# some_value = <your_value>",
            "",
            "# Integer",
            "# some_int = <your_value>",
            "",
            "[test.test2]",
            "  # String",
            "  # some_value2 = <your_value>",
            "  [test.test2.test3]",
            "",
            "[test.test2a]",
            "  # String",
            "  # some_value2a = <your_value>",
            "",
        ]
        config_lines = example._create_example_lines(config_group, "toml")
        self.assertEqual(config_lines, expected_lines)



# Copyright (C) 2019 Catalyst Cloud Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import TestCase

from confspirator import exceptions
from confspirator.tests import utils
from confspirator import groups


DEFAULT_CONF = groups.GroupNamespace(
    "test",
    values={
        "list_val": [1, 2, 3, 4],
        "dict_val": {"one": 1, "two": 2},
        "int_val": 5,
        "str_val": "stuff",
        "group_val": groups.GroupNamespace("group_val", values={"nested_val": 16}),
    },
)


@utils.modify_conf(
    DEFAULT_CONF, {"test.list_val": [{"operation": "append", "value": 5}]}
)
class ModifyConfClassTests(TestCase):
    """test that the class overriding works"""

    def test_list_val(self):
        self.assertTrue(5 in DEFAULT_CONF.list_val)


class ModifyConfFunctionTests(TestCase):
    """Test that function overriding works"""

    def test_list_val_1(self):
        self.assertFalse(5 in DEFAULT_CONF.list_val)

    @utils.modify_conf(
        DEFAULT_CONF, {"test.list_val": [{"operation": "append", "value": 5}]}
    )
    def test_list_val_2(self):
        self.assertTrue(5 in DEFAULT_CONF.list_val)

    def test_list_val_3(self):
        self.assertFalse(5 in DEFAULT_CONF.list_val)


class ModifyConfContextTests(TestCase):
    """test that 'with' context works

    And here we also test that each operation and conf type
    works since this is the easiest way to test it.
    """

    def test_int_val(self):
        self.assertEqual(DEFAULT_CONF.int_val, 5)
        with utils.modify_conf(
            DEFAULT_CONF, {"test.int_val": [{"operation": "override", "value": 6}]}
        ):
            self.assertEqual(DEFAULT_CONF.int_val, 6)
        self.assertEqual(DEFAULT_CONF.int_val, 5)

    def test_list_val_append(self):
        self.assertFalse(5 in DEFAULT_CONF.list_val)
        with utils.modify_conf(
            DEFAULT_CONF, {"test.list_val": [{"operation": "append", "value": 5}]}
        ):
            self.assertTrue(5 in DEFAULT_CONF.list_val)
            self.assertEqual(5, DEFAULT_CONF.list_val[-1])
        self.assertFalse(5 in DEFAULT_CONF.list_val)

    def test_list_val_prepend(self):
        self.assertFalse(5 in DEFAULT_CONF.list_val)
        with utils.modify_conf(
            DEFAULT_CONF, {"test.list_val": [{"operation": "prepend", "value": 5}]}
        ):
            self.assertTrue(5 in DEFAULT_CONF.list_val)
            self.assertEqual(5, DEFAULT_CONF.list_val[0])
        self.assertFalse(5 in DEFAULT_CONF.list_val)

    def test_list_val_prepend_and_append(self):
        self.assertFalse(5 in DEFAULT_CONF.list_val)
        with utils.modify_conf(
            DEFAULT_CONF,
            {
                "test.list_val": [
                    {"operation": "prepend", "value": 0},
                    {"operation": "append", "value": 5},
                    {"operation": "append", "value": 6},
                ]
            },
        ):
            self.assertTrue(0 in DEFAULT_CONF.list_val)
            self.assertTrue(5 in DEFAULT_CONF.list_val)
            self.assertTrue(6 in DEFAULT_CONF.list_val)
            self.assertEqual(0, DEFAULT_CONF.list_val[0])
            self.assertEqual(5, DEFAULT_CONF.list_val[-2])
            self.assertEqual(6, DEFAULT_CONF.list_val[-1])
        self.assertFalse(0 in DEFAULT_CONF.list_val)
        self.assertFalse(5 in DEFAULT_CONF.list_val)
        self.assertFalse(6 in DEFAULT_CONF.list_val)

    def test_list_val_remove(self):
        self.assertTrue(4 in DEFAULT_CONF.list_val)
        with utils.modify_conf(
            DEFAULT_CONF, {"test.list_val": [{"operation": "remove", "value": 4}]}
        ):
            self.assertFalse(4 in DEFAULT_CONF.list_val)
        self.assertTrue(4 in DEFAULT_CONF.list_val)

    def test_dict_val_update(self):
        self.assertFalse("three" in DEFAULT_CONF.dict_val)
        with utils.modify_conf(
            DEFAULT_CONF,
            {"test.dict_val": [{"operation": "update", "value": {"three": 3}}]},
        ):
            self.assertTrue("three" in DEFAULT_CONF.dict_val)
        self.assertFalse("three" in DEFAULT_CONF.dict_val)

    def test_dict_val_delete(self):
        self.assertTrue("two" in DEFAULT_CONF.dict_val)
        with utils.modify_conf(
            DEFAULT_CONF, {"test.dict_val": [{"operation": "delete", "value": "two"}]}
        ):
            self.assertFalse("two" in DEFAULT_CONF.dict_val)
        self.assertTrue("two" in DEFAULT_CONF.dict_val)

    def test_dict_val_overlay(self):
        self.assertFalse("three" in DEFAULT_CONF.dict_val)
        self.assertEqual(2, DEFAULT_CONF.dict_val["two"])
        with utils.modify_conf(
            DEFAULT_CONF,
            {
                "test.dict_val": [
                    {"operation": "overlay", "value": {"three": 3, "two": 5}}
                ]
            },
        ):
            self.assertTrue("three" in DEFAULT_CONF.dict_val)
            self.assertEqual(5, DEFAULT_CONF.dict_val["two"])
        self.assertFalse("three" in DEFAULT_CONF.dict_val)
        self.assertEqual(2, DEFAULT_CONF.dict_val["two"])

    def test_group_val_overlay(self):
        self.assertFalse("three" in DEFAULT_CONF.group_val)
        self.assertEqual(16, DEFAULT_CONF.group_val["nested_val"])
        with utils.modify_conf(
            DEFAULT_CONF,
            {
                "test.group_val": [
                    {"operation": "overlay", "value": {"three": 3, "nested_val": 5}}
                ]
            },
        ):
            self.assertTrue("three" in DEFAULT_CONF.group_val)
            self.assertEqual(5, DEFAULT_CONF.group_val["nested_val"])
        self.assertFalse("three" in DEFAULT_CONF.group_val)
        self.assertEqual(16, DEFAULT_CONF.group_val["nested_val"])

    def test_group_invalid_op(self):
        with self.assertRaises(exceptions.InvalidTestOperation):
            with utils.modify_conf(
                DEFAULT_CONF,
                {"test.group_val": [{"operation": "update", "value": {"three": 3}}]},
            ):
                pass
